#include "scnhacking.h"

#ifdef _MSC_VER
static int strncasecmp(char *s1, char *s2, size_t n) {
  if (n == 0) return 0;

  while (n-- != 0 && tolower(*s1) == tolower(*s2)) {
    if (n == 0 || *s1 == '\0' || *s2 == '\0') break;
    s1++;
    s2++;
  }

  return tolower(*(unsigned char *) s1) - tolower(*(unsigned char *) s2);
}
#endif

entclass_t **alloc_entclasses(unsigned int n) {
  return (entclass_t **) alloc_lots(n, "entclass", sizeof(entclass_t *), (void *(*)()) alloc_entclass);
}

entclass_t **alloc_more_entclasses(unsigned int n, unsigned int num_entclasss, entclass_t **entclasss) {
  return (entclass_t **) alloc_more(n, "entclass", sizeof(entclass_t *), (void *(*)()) alloc_entclass, num_entclasss, (void **) entclasss, sizeof(entclass_t));
}

entclass_t *alloc_entclass() {
  return (entclass_t *) alloc_one("entclass", sizeof(entclass_t));
}

/* can't use alloc_lots() as we don't want to alloc space for the members */
entity_t **alloc_addresses(unsigned int n) {
  unsigned int i;
  entity_t **ret;

  ret = (entity_t **) malloc(n * sizeof(entity_t *));
  TRACE2("alloc_addresses(%u * %u): %x\n", sizeof(entity_t *), n, ret);
  if (! ret) {
    ERROR("out of memory for **addresses", ret);
    return 0;
  }
  for (i = 0; i < n; i++) ret[i] = 0;

  return ret;
}

entity_t **alloc_more_addresses(unsigned int n, unsigned int num_addresses, entity_t **addresses) {
  unsigned int i;
  entity_t **ret;

  ret = (entity_t **) malloc((n + num_addresses) * sizeof(entity_t *));
  TRACE2("alloc_addresses(%u * %u): %x\n", sizeof(entity_t *), n + num_addresses, ret);
  if (! ret) {
    ERROR("out of memory for **addresses", ret);
    return 0;
  }

  for (i = 0; i < num_addresses; i++) ret[i] = addresses[i];
  for (; i < n + num_addresses; i++) ret[i] = 0;
  free(addresses);

  return ret;
}

void cleanup_addresses(unsigned int num_addresses, entity_t **addresses) {
  if (! num_addresses) return;

  if (! addresses) {
    ERROR("no addresses");
    return;
  }

  TRACE2("freeing addresses at %p\n", addresses);
  free(addresses);
}

void cleanup_entclasses(unsigned int num_entclasses, entclass_t **entclasses) {
  unsigned int i;

  if (! num_entclasses) return;

  if (! entclasses) {
    ERROR("no entclasses!\n");
    return;
  }

  for (i = 0; i < num_entclasses; i++) cleanup_entclass(entclasses[i]);
  TRACE2("freeing entclasses at %p\n", entclasses);
  free(entclasses);
}

void cleanup_entclass(entclass_t *entclass) {
  if (! entclass) {
    ERROR("no entclass!\n");
    return;
  }

  if (entclass->classname) free(entclass->classname);
  if (entclass->num) free(entclass->addresses);
  free(entclass);
}

entclass_t *get_entclass(unsigned int num_entclasses, entclass_t **entclasses, char *classname) {
  unsigned int i;
  entclass_t *entclass;

  if (! num_entclasses) return 0;

  if (! entclasses) {
    ERROR("no entclasses!\n");
    return 0;
  }
  if (! classname) {
    ERROR("no classname!\n");
    return 0;
  }

  for (i = 0; i < num_entclasses; i++) {
    entclass = entclasses[i];
    if (str_equiv(classname, entclass->classname)) return entclass;
  }

  DEBUG("no entclass of type \"%s\" found in %u entclass%s\n", classname, num_entclasses, (num_entclasses == 1) ? "" : "es");
  return 0;
}

int sort_entclasses(unsigned int num_entclasses, entclass_t **entclasses) {
  if (! num_entclasses) {
    DEBUG("no entclasses to sort\n");
    return 0;
  }
  if (! entclasses) {
    ERROR("no entclasses!\n");
    return 1;
  }

  qsort((void *) entclasses, (size_t) num_entclasses, sizeof(entclass_t *), sort_entclasses_function);

  return 0;
}

int sort_entclasses_function(const void *a, const void *b) {
  entclass_t *one, *two;

  if (! a || ! b) {
    ERROR("null argument(s)\n");
    return 0;
  }

  one = *(entclass_t **) a;
  two = *(entclass_t **) b;

  TRACE("one: %p <%u>\n", one, one->classname);
  TRACE("two: %p <%u>\n", two, two->classname);

  return strncasecmp(one->classname, two->classname, strlen(one->classname));
}

int summarise_entclasses(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("Unique entity classes: %u\n", scn->num_entclasses);

  return 0;
}

int describe_entclasses(scn_t *scn) {
  unsigned int i;
  entclass_t *entclass;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("Unique entity classes: %u\n", scn->num_entclasses);
  if (! scn->num_entclasses) return 0;

  for (i = 0; i < scn->num_entclasses; i++) {
    entclass = scn->entclasses[i];
    SAY("  Class \"%s\": %u entit%s\n", entclass->classname, entclass->num, (entclass->num == 1) ? "y" : "ies");
  }

  return 0;
}
