#ifndef ENTITY_H
#define ENTITY_H

#define ENTITIES_NUM_OFFSET 0x0c
#define ENTITIES_START_OFFSET 0x34
#define ENTITIES_START_OFFSET_ALSO 0x28

#define CLASSNAME "classname"
#define TARGETNAME "targetname"

struct entity {
  off_t offset;
  unsigned int id;
  unsigned int num_fields;
  int unknown; /* assume magic 0xffffffff */
  struct pair **fields;
};
typedef struct entity entity_t;

struct entity_search_set {
  struct entity_search_set *next;
  int entity_id;
  char *key;
  char *value;
};
typedef struct entity_search_set entity_search_set_t;

struct entity_edit_set {
  struct entity_edit_set *next;
  char *entity_arg;
  int entity_id;
  char *key;
  char *value;
  int options;
};
typedef struct entity_edit_set entity_edit_set_t;

struct entity_add_set {
  struct entity_add_set *next;
  char *name;
  char *classname;
};
typedef struct entity_add_set entity_add_set_t;

struct entity_delete_set {
  struct entity_delete_set *next;
  char *entity_arg;
  int entity_id;
};
typedef struct entity_delete_set entity_delete_set_t;

entity_t **alloc_entities(unsigned int);
entity_t *alloc_entity(void);
void cleanup_entities(thing_t *);
void cleanup_entity(entity_t *);
int find_num_entities(scn_t *);
int find_entities(scn_t *);
int parse_entity(scn_t *, entity_t *, unsigned int, unsigned int *);
int classify_entity(scn_t *, entity_t *);
int write_entities(scn_t *);
char *get_entity_value(entity_t *, char *);
entity_search_set_t *alloc_entity_search_set(void);
int add_to_entity_search_set(entity_search_set_t **, char *, char *, char *);
void cleanup_entity_search_sets(entity_search_set_t *);
int search_entities(scn_t *);
int check_entity(entity_t *, entity_search_set_t *);
entity_edit_set_t *alloc_entity_edit_set(void);
void cleanup_entity_edit_sets(entity_edit_set_t *);
int add_to_entity_edit_set(entity_edit_set_t **, char *, char *, char *, int);
entity_add_set_t *alloc_entity_add_set(void);
void cleanup_entity_add_sets(entity_add_set_t *);
int add_to_entity_add_set(entity_add_set_t **, char *, char *);
entity_delete_set_t *alloc_entity_delete_set(void);
void cleanup_entity_delete_sets(entity_delete_set_t *);
int add_to_entity_delete_set(entity_delete_set_t **, char *);
int edit_entities(scn_t *);
int edit_entity(scn_t *, entity_edit_set_t *);
int add_entities(scn_t *);
int add_entity(scn_t *, entity_add_set_t *);
int delete_entities(scn_t *);
int delete_entity(scn_t *, entity_delete_set_t *);
int summarise_entities(scn_t *);
int describe_entities(scn_t *);
int describe_entity(entity_t *, int);

#endif
