#include "scnhacking.h"

#define NUM_LIGHT_ENTITIES 3
static char *light_entities[NUM_LIGHT_ENTITIES] = {
  "light", "light_ambient", "light_spot"
};

colour_replace_set_t *alloc_colour_replace_set() {
  return (colour_replace_set_t *) alloc_one("colour_replace_set", sizeof(colour_replace_set_t));
}

void cleanup_colour_replace_sets(colour_replace_set_t *sets) {
  TRACE2("freeing colour_replace_sets at %p\n", sets);
  gll_free((gll_t *) sets);
}

int add_to_colour_replace_set(colour_replace_set_t **sets, char *light_arg, char *red_arg, char *green_arg, char *blue_arg) {
  colour_replace_set_t *last, *tmp;
  char *bogus;
  long long_value;
  double double_value;

  tmp = alloc_colour_replace_set();
  if (! tmp) {
    ERROR("out of memory for colour_replace_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  /* light ID to replace */
  if (str_equiv(light_arg, "all")) {
    tmp->first_light_id = 0;
    tmp->last_light_id = 0xffffffff;
    TRACE("adjusting all lights\n");
  }
  else {
    long_value = strtol(light_arg, &bogus, 0);
    if (*bogus) {
      ERROR("bogus light ID \"%s\" - use a number or \"all\"\n", light_arg);
      cleanup_colour_replace_sets(tmp);
      return 2;
    }
    tmp->first_light_id = tmp->last_light_id = (unsigned int) long_value;
  }

  /* parse red values */
  long_value = strtol(red_arg, &bogus, 0);
  if (*bogus) {
    /* not an integer => try float */
    double_value = strtod(red_arg, &bogus);
    if (*bogus) {
      ERROR("bogus red value \"%s\" - use 0.0-1.0 or 0-255\n", red_arg);
      cleanup_colour_replace_sets(tmp);
      return 3;
    }
    if (double_value < 0 || double_value > 1.0) {
      ERROR("red value %0.4f out of range - use 0.0 to 1.0\n", (float) double_value);
      cleanup_colour_replace_sets(tmp);
      return 4;
    }
    tmp->red = (float) double_value;
  }
  else {
    if (long_value < 0 || long_value > 255) {
      ERROR("red value %ld out of range - use 0 to 255\n", long_value);
      cleanup_colour_replace_sets(tmp);
      return 5;
    }
    tmp->red = (float) (long_value / 255.0);
  }
  TRACE("tmp->red: %.4f\n", tmp->red);

  /* parse green values */
  long_value = strtol(green_arg, &bogus, 0);
  if (*bogus) {
    /* not an integer => try float */
    double_value = strtod(green_arg, &bogus);
    if (*bogus) {
      ERROR("bogus green value \"%s\" - use 0.0-1.0 or 0-255\n", green_arg);
      cleanup_colour_replace_sets(tmp);
      return 6;
    }
    if (double_value < 0 || double_value > 1.0) {
      ERROR("green value %0.4f out of range - use 0.0 to 1.0\n", (float) double_value);
      cleanup_colour_replace_sets(tmp);
      return 7;
    }
    tmp->green = (float) double_value;
  }
  else {
    if (long_value < 0 || long_value > 255) {
      ERROR("green value %ld out of range - use 0 to 255\n", long_value);
      cleanup_colour_replace_sets(tmp);
      return 8;
    }
    tmp->green = (float) (long_value / 255.0);
  }
  TRACE("tmp->green: %.4f\n", tmp->green);

  /* parse blue values */
  long_value = strtol(blue_arg, &bogus, 0);
  if (*bogus) {
    /* not an integer => try float */
    double_value = strtod(blue_arg, &bogus);
    if (*bogus) {
      ERROR("bogus blue value \"%s\" - use 0.0-1.0 or 0-255\n", blue_arg);
      cleanup_colour_replace_sets(tmp);
      return 9;
    }
    if (double_value < 0 || double_value > 1.0) {
      ERROR("blue value %0.4f out of range - use 0.0 to 1.0\n", (float) double_value);
      cleanup_colour_replace_sets(tmp);
      return 10;
    }
    tmp->blue = (float) double_value;
  }
  else {
    if (long_value < 0 || long_value > 255) {
      ERROR("blue value %ld out of range - use 0 to 255\n", long_value);
      cleanup_colour_replace_sets(tmp);
      return 11;
    }
    tmp->blue = (float) (long_value / 255.0);
  }
  TRACE("tmp->blue: %.4f\n", tmp->red);

  last  = (colour_replace_set_t *) gll_last((gll_t *) *sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *sets = tmp;
  /*gll_debug((gll_t *) *sets);*/

  return 0;
}

/* walk through the file and set the RGB lighting values */
int replace_colours(scn_t *scn) {
  colour_replace_set_t *set;
  int ret;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("setting light colours\n");
  GLL_LOOP(set, scn->option->colour_replaces) {
    ret = replace_colour(scn, set);
    if (ret) {
      ERROR("replace_colour() returned %d\n", ret);
      return 2;
    }
  }

  return 0;
}

int replace_colour(scn_t *scn, colour_replace_set_t *set) {
  unsigned short i, j, len;
  pair_t *pair, *new_pair, **new_pairs;
  thing_t *entities;
  entity_t *entity;
  char *string;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }
  if (! set) {
    ERROR("no set!\n");
    return 2;
  }

  /* parse numbers if necessary */
  if (set->last_light_id = 0xffffffff) {
    set->last_light_id = scn->lightmaps.num_lights - 1;
  }

  /* sanity */
  if (set->last_light_id < set->first_light_id) {
    ERROR("last light ID < first light ID!\n");
    return 3;
  }
  if (set->last_light_id >= scn->lightmaps.num_lights) {
    ERROR("last light ID %u out of range (%u light%s)\n", set->last_light_id, scn->lightmaps.num_lights, (scn->lightmaps.num_lights == 1) ? "" : "s");
    return 4;
  }

  entities = &scn->entities;

  for (i = set->first_light_id; i <= set->last_light_id; i++) {
    /* get the entity for this light */
    entity = (entity_t *) scn->lightmaps.addresses[i];
    TRACE("processing light %u which is entity at %p\n", i, entity);

    /* get the pair with the existing colour */
    pair = 0;
    for (j = 0; j < entity->num_fields; j++) {
      new_pair = entity->fields[j];
      TRACE1("checking key \"%s\" for string \"%s\"\n", new_pair->key, COLOUR);
      if (str_equiv(new_pair->key, COLOUR)) {
        pair = new_pair;
        break;
      }
    }
    /* no Color field => some lights don't have it */
    if (! pair) {
      if (! scn->option->force_colours) {
        SAY("ignoring light %u which has no %s field\n", i, COLOUR);
        continue;
      }
      SAY("creating %s field for light %u\n", COLOUR, i);
      new_pairs = alloc_more_pairs(1, entity->num_fields, entity->fields);
      if (! new_pairs) {
        ERROR("out of memory for new field");
        return 5;
      }
      entity->fields = new_pairs;
      pair = entity->fields[entity->num_fields++];

      /* set key for new pair */
      pair->key = alloc_name(COLOUR, &len);
      if (! len) {
        ERROR("out of memory for new field key");
        return 6;
      }
      pair->key_length = len;

      entities->length += pair->key_length + pair->value_length + 4;
    }

    /* allocate space for new value */
    string = (char *) malloc(18);
    if (! string) {
      ERROR("out of memory for new value");
      return 7;
    }

    /* set new value */
    if (snprintf(string, 18, "%0.3f %0.3f %0.3f", set->red, set->green, set->blue) == -1) {
      ERROR("snprintf() barfed\n");
      free(string);
      return 8;
    }
    if (pair->value) free(pair->value);
    pair->value = string;
    entities->length -= (pair->value_length - 18);
    pair->value_length = 18;
    TRACE("now value is %s\n", pair->value);
  }

  return 0;
}

int is_light(entity_t *entity) {
  unsigned short i, j;
  pair_t *pair;

  if (! entity) {
    ERROR("no entity!\n");
    return -1;
  }

  for (i = 0; i < entity->num_fields; i++) {
    pair = entity->fields[i];
    if (! str_equiv(pair->key, CLASSNAME)) continue;
    for (j = 0; j < NUM_LIGHT_ENTITIES; j++) {
      if (str_equiv(pair->value, light_entities[j])) return 1;
    }
  }

  return 0;
}

int summarise_lights(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  SAY("Lights: %u\n", scn->lightmaps.num_lights);

  return 0;
}

int describe_lights(scn_t *scn) {
  unsigned short i;
  entity_t *entity;

  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  if (scn->lightmaps.num_lights) SAY("Lights: %u\n", scn->lightmaps.num_lights);
  for (i = 0; i < scn->lightmaps.num_lights; i++) {
    entity = (entity_t *) scn->lightmaps.addresses[i];
    SAY("  Light %u is ", i);
    describe_entity(entity, 4);
  }

  return 0;
}

int count_lights(scn_t *scn) {
  entity_t *entity;
  unsigned int i, j;

  scn->lightmaps.num_lights = 0;
  cleanup_lightmaps(&scn->lightmaps);

  DEBUG("counting lights\n");
  j = 0;
  for (i = 0; i < scn->entities.num; i++) {
    entity = scn->entities.things[i];
    if (is_light(entity)) {
      TRACE("entity %u is a light\n", i);
      j++;
    }
  }
  DEBUG("there ... are ... %d ... lights!\n", j);
  scn->lightmaps.addresses = alloc_ints(j);
  if (! scn->lightmaps.addresses) {
    ERROR("out of memory for lightmaps");
    return 1;
  }
  scn->lightmaps.num_lights = j;

  j = 0;
  for (i = 0; i < scn->entities.num; i++) {
    entity = scn->entities.things[i];
    if (is_light(entity)) {
      TRACE("light %u at %p\n", j, entity);
      scn->lightmaps.addresses[j++] = (void *) entity;
    }
  }

  return 0;
}
