#include "scnhacking.h"

void cleanup_lightmaps(lightmaps_t *lightmaps) {
  if (! lightmaps) {
    ERROR("no lightmaps!\n");
    return;
  }

  if (lightmaps->addresses) free(lightmaps->addresses);
}

int find_lightmaps(scn_t *scn) {
  if (! scn || ! scn->fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  /* skip to magic offset */
  if (fseek(scn->fd, LIGHTMAPS_OFFSET, SEEK_SET)) {
    ERROR("can't seek to location of lightmaps offset in %s\n", scn->filename);
    return 2;
  }

  /* read start of lightmaps */
  if (read_uint32(scn->fd, (unsigned int *) &scn->lightmaps.offset)) {
    ERROR("can't read lightmaps offset in %s\n", scn->filename);
    return 3;
  }
  /* read number of lights */
  if (read_uint32(scn->fd, &scn->lightmaps.num_lights)) {
    ERROR("can't read number of lights in %s\n", scn->filename);
    return 4;
  }

  DEBUG("%s: %d lights starting at 0x%08x\n", scn->filename, scn->lightmaps.num_lights, scn->lightmaps.offset);

  /* allocate entity pointers */
  if (! scn->lightmaps.num_lights) return 0;

  scn->lightmaps.addresses = (unsigned int **) malloc(scn->lightmaps.num_lights * sizeof(unsigned int *));
  if (! scn->lightmaps.addresses) {
    ERROR("out of memory for addresses");
    return 5;
  }

  return 0;
}

/*
  this would actually do something, here we just check whether the lightmaps 
  exist or not by abusing the length field of the lightmaps struct
*/
int parse_lightmaps(scn_t *scn) {
  char buf;

  if (! scn || ! scn->fd) {
    ERROR("no SCN!\n");
    return 1;
  }

  /* skip to magic offset */
  DEBUG("checking if there is actually stuff at start of lightmaps\n");
  if (fseek(scn->fd, scn->lightmaps.offset, SEEK_SET)) {
    ERROR("can't seek to start of lightmaps in %s", scn->filename);
    return 2;
  }

  /* try to read something */
  if (read_char(scn->fd, &buf)) {
    if (feof(scn->fd)) {
      DEBUG("no lightmaps\n");
      scn->lightmaps.length = 0;
      return 0;
    }
    ERROR("can't check for lightmaps in %s", scn->filename);
    return 3;
  }

  DEBUG("yes there are lightmaps at 0x%08x\n", scn->lightmaps.offset);
  scn->lightmaps.length = 1; /* XXX: set to something useful */

  return 0;
}

int strip_lightmaps(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  if (! scn->lightmaps.length) {
    ERROR("no lightmaps to strip\n");
    return 2;
  }

  /* nothing to do */
  SAY("stripping lightmaps from %s\n", scn->filename);

  return 0;
}

int summarise_lightmaps(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  if (scn->lightmaps.length) {
    SAY("Lightmaps at 0x%08x\n", scn->lightmaps.offset);
  }
  else {
    SAY("No lightmaps\n");
  }

  return 0;
}

int describe_lightmaps(scn_t *scn) {
  if (! scn) {
    ERROR("no SCN!\n");
    return 1;
  }

  if (scn->lightmaps.length) {
    SAY("Lightmaps at 0x%08x\n", scn->lightmaps.offset);
  }
  else {
    SAY("No lightmaps\n");
  }

  return 0;
}
