#include "scnhacking.h"

#define NUM_DESCRIPTIONS 3
static char *description_names[NUM_DESCRIPTIONS] = {
  "entities", "lights", "classes"
};
static unsigned int description_options[NUM_DESCRIPTIONS] = {
  DESCRIBE_ENTITIES
 ,DESCRIBE_LIGHTS
 ,DESCRIBE_CLASSES
};
static struct description_types {
  char *name;
  unsigned int option;
} descriptions[NUM_DESCRIPTIONS];

extern char **filenames;
option_t **alloc_options(unsigned short n) {
  return (option_t **) alloc_lots(n, "option", sizeof(option_t *), (void *(*)()) alloc_option);
}

option_t *alloc_option() {
  option_t *ret;

  ret = (option_t *) alloc_one("option", sizeof(option_t));
  if (! ret) return 0;
  initialise_option(ret);

  return ret;
}

void cleanup_option(option_t *option) {
  if (! option) {
    DEBUG("null option pointer\n");
    return;
  }
  if (--option->scncount) {
    DEBUG("option at %p still in use; scncount: %u\n", option, option->scncount);
    return;
  }
  if (option->colour_replaces) {
    cleanup_colour_replace_sets(option->colour_replaces);
  }
  if (option->entity_searches) {
    cleanup_entity_search_sets(option->entity_searches);
  }
  if (option->entity_adds) {
    cleanup_entity_add_sets(option->entity_adds);
  }
  if (option->entity_deletes) {
    cleanup_entity_delete_sets(option->entity_deletes);
  }
  if (option->entity_edits) {
    cleanup_entity_edit_sets(option->entity_edits);
  }
  TRACE2("freeing option at %p\n", option);
  free(option);
}

void initialise_option(option_t *option) {
  memset(option, 0, sizeof(option_t));
}

void cleanup_options(unsigned short num_options, option_t **options) {
  unsigned short i;
  for (i = 0; i < num_options; i++) cleanup_option(options[i]);
  free(options);
}

#define skip(x) \
  { for (j = i; j < argc - x; j++) argv[j] = argv[j + x]; argc -= x; i--; }
struct scn **process_options(int *_argc, char **_argv, int *num_scns) {
  struct scn **scns;
  struct option **options, option;
  int i, j, num_skip;
  unsigned short len;
  int cur;
  int argc;
  char **argv;
  int last_option;
  int num_filegroups;

  /* this is fucking ugly */
  for (cur = 0; cur < NUM_DESCRIPTIONS; cur++) {
    descriptions[cur].name = description_names[cur];
    descriptions[cur].option = description_options[cur];
  }

  *num_scns = num_filegroups = 0;

  /* must have at least one scn */
  argc = *_argc - 1;
  if (! argc) {
    usage();
    return 0;
  }
  argv = _argv + 1;

  /* worst case scenario is all arguments are option groups */
  options = alloc_options(argc);
  if (! options) {
    ERROR("out of memory for options");
    *num_scns = -1;
    return 0;
  }

  /* or, worst case, all are files */
  scns = alloc_scns(argc);
  if (! scns) {
    ERROR("out of memory for scns");
    *num_scns = -1;
    return 0;
  }

  cur = 0;
  last_option = OPTION_NONE;
  for (i = 0; i < argc; i++) {
    initialise_option(&option);
    num_skip = parse_option(&option, argc, argv, i);
    if (num_skip < 0) {
      for (i = 0; i < *_argc - 1; i++) {
        cleanup_option(options[i]);
        cleanup_scn(scns[i]);
      }
      free(options);
      free(scns);
      if (num_skip % 2) (void) usage();
      return 0;
    }
    if (num_skip) {
      /* argument */
      if (last_option == OPTION_FILENAME && options[cur]->number) {
        cur++;
        DEBUG("new option group with \"%s\": %d\n", argv[i], cur);
      }
      last_option = OPTION_ARGUMENT;
      copy_options(options[cur], &option);
      if (num_skip) skip(num_skip);
    }
    else {
      /* filename */
      DEBUG("option not recognised, treating as file: %s\n", argv[i]);
      scns[*num_scns]->filename = alloc_name(argv[i], &len);
      if (! len) {
        ERROR("out of memory for file %d name\n", *num_scns);
        *num_scns = -1;
        return 0;
      }
      scns[*num_scns]->option = options[cur];
      scns[*num_scns]->option->scncount++;
      DEBUG("scn %d filename %s at %p has option %d at %p (scncount=%u)\n", *num_scns, scns[*num_scns]->filename, scns[*num_scns]->filename, cur, options[cur], scns[*num_scns]->option->scncount);
      ++*num_scns;
      if (last_option != OPTION_FILENAME) num_filegroups++;

      last_option = OPTION_FILENAME;
    }
  }

  /* this isn't BSD */
  DEBUG("current option: %d, num_filegroups: %d\n", cur, num_filegroups);
  if (num_filegroups && cur >= num_filegroups) {
    DEBUG("handling trailing options (cur: %d, num_filegroups: %d)\n", cur, num_filegroups);
    for (i = num_filegroups; i <= cur; i++) {
      DEBUG("appending option group %d to group %d\n", i, num_filegroups - 1);
      copy_options(options[num_filegroups - 1], options[i]);
    }
  }

  /* free unused options and scns */
  argc = *_argc - 1;
  if (argc == *num_scns) cur = 1;
  else cur = argc - *num_scns;
  DEBUG("cur: %d, *num_scns: %d, num_filegroups: %d, argc: %d\n", cur, *num_scns, num_filegroups, argc);
  for (i = cur; i < argc; i++) {
    if (options[i]->number) {
      ERROR("please tell furrycat he's a tosser (%d)\n", i);
    }
    DEBUG("freeing unused option %d at %p with number %u\n", i, options[i], options[i]->number);
    free(options[i]);
  }
  free(options);

  for (i = *num_scns; i < argc; i++) {
    if (scns[i]->filename) {
      ERROR("please tell furrycat he's a wanker (%d)\n", i);
    }
    DEBUG("freeing unused SCN %d at %p\n", i, scns[i]);
    free(scns[i]);
  }

  /* no filenames => bad */
  if (! num_filegroups) {
    (void) usage();
    return 0;
  }

  return scns;
}

#if 0
scn_t **process_options(int *_argc, char **_argv, int *num_scns) {
  scn_t **scns;
  option_t **options, option;
  int i, j, num_skip;
  unsigned short len;
  int cur;
  int argc;
  char **argv;
  int last_option;
  int num_filegroups, num_options;

  /* this is fucking ugly */
  DESCRIBE_ALL = 0;
  for (cur = 0; cur < NUM_DESCRIPTIONS; cur++) {
    descriptions[cur].name = description_names[cur];
    descriptions[cur].option = description_options[cur];
    DESCRIBE_ALL += description_options[cur];
  }

  *num_scns = num_filegroups = 0;

  /* must have at least one scn */
  argc = *_argc - 1;
  if (! argc) {
    usage();
    return 0;
  }
  argv = _argv + 1;

  /* worst case scenario is all arguments are option groups */
  num_options = argc;
  options = alloc_options(num_options);
  if (! options) {
    ERROR("out of memory for options");
    *num_scns = -1;
    return 0;
  }

  /* or, worst case, all are files */
  scns = alloc_scns(argc);
  if (! scns) {
    ERROR("out of memory for scns");
    *num_scns = -1;
    cleanup_options(num_options, options);
    return 0;
  }

  cur = 0;
  last_option = OPTION_NONE;
  num_options;
  for (i = 0; i < argc; i++) {
    initialise_option(&option);
    num_skip = parse_option(&option, argc, argv, i);
    if (num_skip < 0) {
      for (i = 0; i < *_argc - 1; i++) {
        cleanup_option(options[i]);
        cleanup_scn(scns[i]);
      }
      free(options);
      free(scns);
      if (num_skip % 2) (void) usage();
      return 0;
    }
    if (num_skip) {
      /* argument */
      if (last_option == OPTION_FILENAME && options[cur]->number) {
        cur++;
        DEBUG("new option group with \"%s\": %d\n", argv[i], cur);
      }
      last_option = OPTION_ARGUMENT;
      copy_options(options[cur], &option);
      if (num_skip) skip(num_skip);
    }
    else {
      /* filename */
      DEBUG("option not recognised, treating as file: %s\n", argv[i]);
      scns[*num_scns]->filename = alloc_name(argv[i], &len);
      if (! len) {
        ERROR("out of memory for file %d name\n", *num_scns);
        *num_scns = -1;
        cleanup_options(num_options, options);
        return 0;
      }
      scns[*num_scns]->option = options[cur];
      scns[*num_scns]->option->scncount++;
      DEBUG("scn %d filename %s at %p has option %d at %p (scncount=%u)\n", *num_scns, scns[*num_scns]->filename, scns[*num_scns]->filename, cur, options[cur], scns[*num_scns]->option->scncount);
      ++*num_scns;
      if (last_option != OPTION_FILENAME) num_filegroups++;

      last_option = OPTION_FILENAME;
    }
  }

  /* this isn't BSD */
  DEBUG("current option: %d, num_filegroups: %d\n", cur, num_filegroups);
  if (num_filegroups && cur >= num_filegroups) {
    DEBUG("handling trailing options (cur: %d, num_filegroups: %d)\n", cur, num_filegroups);
    for (i = num_filegroups; i <= cur; i++) {
      DEBUG("appending option group %d to group %d\n", i, num_filegroups - 1);
      copy_options(options[num_filegroups - 1], options[i]);
    }
  }

  /* free unused options and scns */
  argc = *_argc - 1;
  if (argc == *num_scns) cur = 1;
  else cur = argc - *num_scns;
  DEBUG("cur: %d, *num_scns: %d, num_filegroups: %d, argc: %d\n", cur, *num_scns, num_filegroups, argc);
#if 0
  for (i = cur; i < argc; i++) {
    if (options[i]->number) {
      ERROR("please tell furrycat he's a tosser (%d)\n", i);
    }
    DEBUG("freeing unused option %d at %p with number %u\n", i, options[i], options[i]->number);
    free(options[i]);
  }
#endif
  for (i = 0; i < num_options; i++) {
    if (options[i]->number) {
      ERROR("please tell furrycat he's a wanker (%d)\n", i);
    }
    free(options[i]);
  }
  free(options);

  for (i = *num_scns; i < argc; i++) {
    if (scns[i]->filename) {
      ERROR("please tell furrycat he's a wanker (%d)\n", i);
    }
    DEBUG("freeing unused SCN %d at %p\n", i, scns[i]);
    free(scns[i]);
  }

  /* no filenames => bad */
  if (! num_filegroups) {
    (void) usage();
    return 0;
  }

  return scns;
}
#endif

int usage() {
  error_dbg("usage: scnhacking [options] filenames [[options] [filenames]]\n");
  error_dbg("options: -v,  --verbose                 [be verbose]\n");
#ifdef EBUG
  error_dbg("         -d,  --debug                   [debugging mode]\n");
  error_dbg("         -sd, --super-debug             [hardcore debugging mode]\n");
#endif
  error_dbg("         -q,  --quiet                   [no output at all]\n");
  error_dbg("         -c,  --classify                [show type of SCN]\n");
  error_dbg("              --describe feature        [show detailed description of SCN]\n");
  error_dbg("         -D                             [equivalent to --describe all]\n");
  error_dbg("         -S,  --summary                 [show summary of SCN]\n");
  error_dbg("         -s,  --search-entity key value [show entities with given attribute]\n");
  error_dbg("              --search-entity-id id     [show entity with given ID]\n");
  error_dbg("              --search-entity-class class\n");
  error_dbg("                                        [show entities of given class]\n");
  error_dbg("              --strip-lights            [remove all lightmaps]\n");
  error_dbg("         -r,  --replace-colours colour r g b\n");
  error_dbg("                                        [set RGB values for colour in]\n");
  error_dbg("              --add-attribute entity key value\n");
  error_dbg("                                        [add a new attribute for an entity]\n");
  error_dbg("              --edit-attribute entity key newvalue\n");
  error_dbg("                                        [edit an entity's attribute]\n");
  error_dbg("              --delete-attribute entity key\n");
  error_dbg("                                        [delete an entity's attribute]\n");
  error_dbg("         -ae, --add-entity name classname\n");
  error_dbg("                                        [create an entity of type classname]\n");
  error_dbg("         -de, --delete-entity entity    [delete an entity]\n");
  error_dbg("         -o,  --output filename         [write new file]\n");
  error_dbg("              --outfile filename\n");
  error_dbg("              --prefix prefix\n");
  error_dbg("                                        [add prefix to filename for output]\n");
  error_dbg("notes: --summary is implicit when no options are given\n");
  error_dbg("       all options preceding filenames are applied to those files\n");
  error_dbg("       options following files are applied to the next set of files\n");
  error_dbg("       use \"--outfile -\" to write new SCN to stdout\n");
  error_dbg("       use -d and -v together to get more debugging info\n");
  error_dbg("       if more than one of -d/-v or -q is given, the last takes precedence\n");
  return 1;
}

int help() {
  usage();
  return 1;
}

/* macro to remove processed arguments from the argv list */
/* check if argv[i] is valid and put it in option if necessary */
#define argument(x) str_equals(argv[i], x)
int parse_option(option_t *option, int argc, char **argv, int i) {
  int j, description;
  unsigned short len;

  if (argument("-V") || argument("--version")) {
    exit(version());
  }
  else if (argument("-d") || argument("--debug")) {
    debug = 1; quiet = 0;
    error = stdout;
    DEBUG("entering DEBUG mode\n");
    return 1;
  }
  else if (argument("-h") || argument("--help")) {
    exit(help());
  }
  else if (argument("--strip-lights") || argument("--strip-lmaps") || argument("--strip-lightmaps")) {
    option->strip_lightmaps = 1;
    DEBUG("option --strip-lights\n");
    option->number++;
    return 1;
  }
  else if (argument("-r") || argument("--replace-colours") || argument("--replace-colors") || argument("--replace-colour") || argument("--replace-color")) {
    if (i >= argc - 4) goto replace_colours_syntax;
    if (add_to_colour_replace_set(&option->colour_replaces, argv[i + 1], argv[i + 2], argv[i + 3], argv[i + 4])) {
      ERROR("can't add colour replacement set\n");
replace_colours_syntax:
      error_dbg("--replace-colours syntax: light r g b\n");
      error_dbg("light: light ID to replace or \"all\" for all\n");
      error_dbg("r g b: RGB values for replacement colour\n");
      error_dbg("       use values in the range 0.0-1.0 or 0-255\n");
      return -2;
    }
    DEBUG("replacing colour \"%s\" with RGB (%s, %s, %s)\n", argv[i + 1], argv[i + 2], argv[i + 3], argv[i + 4]);
    option->number++;
    return 5;
  }
  else if (argument("--force-colours") || argument("--force-colors") || argument("--force-colour") || argument("--force-color")) {
    DEBUG("will create %s keys in entities which lack them\n", COLOUR);
    option->force_colours = 1;
    return 1;
  }
  else if (argument("-s") || argument("--search-entity") || argument("--search-entities") || argument("--search-entity-attribute") || argument("--search-entities-attribute") || argument("--search-entity-by-attribute") || argument("--search-entities-by-attribute")) {
    if (i >= argc - 2) goto search_entity_syntax;
    if (add_to_entity_search_set(&option->entity_searches, 0, argv[i + 1], argv[i + 2])) {
search_entity_syntax:
      error_dbg("--search-entity syntax: key value\n");
      error_dbg("key: entity property; eg: ClassName or \"any\" for any\n");
      error_dbg("value: value for property or \"\" for any\n");
      return -2;
    }
    DEBUG("searching for entity %s by %s\n", argv[i + 1], argv[i + 2]);
    option->number++;
    return 3;
  }
  else if (argument("--search-entity-id") || argument("--search-entities-id") || argument("--search-entity-by-id") || argument("--search-entities-by-id") || argument("--search-id")) {
    if (i >= argc - 1) goto search_entity_id_syntax;
    if (add_to_entity_search_set(&option->entity_searches, argv[i + 1], 0, 0)) {
search_entity_id_syntax:
      error_dbg("--search-entity-id syntax: entity\n");
      error_dbg("entity: entity ID\n");
      return -2;
    }
    DEBUG("searching for entity %s\n", argv[i + 1]);
    option->number++;
    return 2;
  }
  else if (argument("--search-entity-class") || argument("--search-entities-class") || argument("--search-entity-by-class") || argument("--search-entities-by-class") || argument("--search-class")) {
    if (i >= argc - 1) goto search_entity_class_syntax;
    if (add_to_entity_search_set(&option->entity_searches, 0, CLASSNAME, argv[i + 1])) {
search_entity_class_syntax:
      error_dbg("--search-entity-class syntax: class\n");
      error_dbg("class: entity class; eg worldspawn\n");
      return -2;
    }
    DEBUG("searching for entity class %s\n", argv[i + 1]);
    option->number++;
    return 2;
  }
  else if (argument("--add-attribute") || argument("--add-entity-attribute")) {
    if (i >= argc - 3) goto add_attribute_syntax;
    if (add_to_entity_edit_set(&option->entity_edits, argv[i + 1], argv[i + 2], argv[i + 3], EDIT_ADD)) {
      ERROR("can't add attribute edit set\n");
add_attribute_syntax:
      error_dbg("--add-attribute syntax: entity key value\n");
      error_dbg("entity: entity ID\n");
      error_dbg("key: entity property; eg: TargetName\n");
      error_dbg("value: value for property\n");
      return -2;
    }
    DEBUG("adding entity \"%s\" attribute: %s=\"%s\"\n", argv[i + 1], argv[i + 2], argv[i + 3]);
    option->number++;
    return 4;
  }
  else if (argument("--delete-attribute") || argument("--delete-entity-attribute")) {
    if (i >= argc - 2) goto delete_attribute_syntax;
    if (add_to_entity_edit_set(&option->entity_edits, argv[i + 1], argv[i + 2], argv[i + 3], EDIT_DELETE)) {
      ERROR("can't add attribute edit set\n");
delete_attribute_syntax:
      error_dbg("--delete-attribute syntax: entity key\n");
      error_dbg("entity: entity ID\n");
      error_dbg("key: entity property; eg: TargetName\n");
      return -2;
    }
    DEBUG("deleting entity \"%s\" attribute: %s\n", argv[i + 1], argv[i + 2]);
    option->number++;
    return 3;
  }
  else if (argument("-e") || argument("--edit-attribute") || argument("--edit-entity-attribute")) {
    if (i >= argc - 3) goto edit_attribute_syntax;
    if (add_to_entity_edit_set(&option->entity_edits, argv[i + 1], argv[i + 2], argv[i + 3], EDIT_EDIT | EDIT_DELETE)) {
      ERROR("can't add attribute edit set\n");
edit_attribute_syntax:
      error_dbg("--edit-attribute syntax: entity key value\n");
      error_dbg("entity: entity ID\n");
      error_dbg("key: entity property; eg: TargetName\n");
      error_dbg("value: value for property\n");
      return -2;
    }
    DEBUG("editing entity \"%s\" attribute: %s=\"%s\"\n", argv[i + 1], argv[i + 2], argv[i + 3]);
    option->number++;
    return 4;
  }
  else if (argument("-ae") || argument("--add-entity")) {
    if (i >= argc - 2) goto add_entity_syntax;
    if (add_to_entity_add_set(&option->entity_adds, argv[i + 1], argv[i + 2])) {
      ERROR("can't add entity add set\n");
add_entity_syntax:
      error_dbg("--add-entity syntax: classname name\n");
      error_dbg("classname: entity class type; eg: cell\n");
      error_dbg("name: entity name\n");
      return -2;
    }
    DEBUG("adding entity \"%s\" of type \"%s\"\n", argv[i + 1], argv[i + 2]);
    option->number++;
    return 3;
  }
  else if (argument("-de") || argument("--delete-entity")) {
    if (i >= argc - 1) goto delete_entity_syntax;
    if (add_to_entity_delete_set(&option->entity_deletes, argv[i + 1])) {
      ERROR("can't add entity delete set\n");
delete_entity_syntax:
      error_dbg("--delete-entity syntax: entity\n");
      error_dbg("entity: entity ID\n");
      return -2;
    }
    DEBUG("deleting entity \"%s\"\n", argv[i + 1]);
    option->number++;
    return 2;
  }
  else if (argument("-o") || argument("--output") || argument("--outfile")) {
    if (i >= argc - 1) return 0;
    if (str_equals(argv[i + 1], "-")) {
      option->fd = stdout;
      quiet = 1;
      return 2;
    }
    else if (option->fd) {
      error_dbg("can't mix --outfile <file> and --outfile -\n");
      return -1;
    }
    option->outfile = alloc_name(argv[i + 1], &len);
    if (! len) {
      ERROR("out of memory for output filename");
      return -1;
    }
    DEBUG("output file is \"%s\"\n", option->outfile);
    return 2;
  }
  else if (argument("-q") || argument("--quiet")) {
    quiet = 1; verbose = debug = 0;
    DEBUG("entering QUIET mode - if you can read this it's a bug!\n");
    return 1;
  }
  else if (argument("-v") || argument("--verbose")) {
    verbose++; quiet = 0;
    DEBUG("entering TRACE mode\n");
    return 1;
  }
  else if (argument("-sd") || argument("--super-debug")) {
    debug = 1; verbose = 999; quiet = 0;
    error = stdout;
    DEBUG("entering extremely verbose debugging mode\n");
    return 1;
  }
  else if (argument("-D")) {
    option->describe |= DESCRIBE_ALL | DESCRIBE_STANDARD;
    DEBUG("will dump extended information about this SCN\n");
    option->number++;
    return 1;
  }
  else if (argument("--describe")) {
    if (i >= argc - 1) goto describe_syntax;
    description = 0;
    for (j = 0; j < NUM_DESCRIPTIONS; j++) {
      if (str_equiv(argv[i + 1], descriptions[j].name)) option->describe |= descriptions[j].option;
      description |= descriptions[j].option;
    }
    if (str_equiv(argv[i + 1], "all")) option->describe |= description;
    option->describe |= DESCRIBE_STANDARD;
    DEBUG("will dump information about this SCN: option %u\n", option->describe);
    option->number++;
    if (option->describe == DESCRIBE_STANDARD) return 1;
    return 2;
describe_syntax:
    error_dbg("--describe syntax: feature\n");
    error_dbg("Features:\n");
    for (j = 0; j < NUM_DESCRIPTIONS; j++) error_dbg("  %s\n", descriptions[j]);
    error_dbg("  all - all features listed above\n");
    return -2;
  }
  else if (argument("-S") || argument("--summary")) {
    if (! option->describe) option->describe |= DESCRIBE_SUMMARY;
    DEBUG("will show summary information about this SCN\n");
    option->number++;
    return 1;
  }
  else if (argument("--prefix")) {
    if (i >= argc - 1) return -1;
    option->prefix = argv[i + 1];
    DEBUG("file prefix: %s\n", option->prefix);
    option->number++;
    return 2;
  }
  else if (argument("--suffix")) {
    if (i >= argc - 1) return -1;
    option->suffix = argv[i + 1];
    DEBUG("file suffix: %s\n", option->suffix);
    option->number++;
    return 2;
  }

  return 0;
}

int version() {
  SAY("scnhacking version %s", VERSION);
#ifdef _MSC_VER
  SAY(" - Visual C++");
#endif
#ifdef __CYGWIN32__
  SAY(" - Cygwin");
#endif
#ifdef EBUG
  SAY(" DEBUG");
#endif
  SAY(" by [furrycat], %s\n", VERSION_DATE);
  return 0;
}

/* append options to option set */
void copy_options(option_t *dest, option_t *src) {
  TRACE("copying option at %p to %p\n", src, dest);
  if (src->number) dest->number++;
  if (src->strip_lightmaps) dest->strip_lightmaps = src->strip_lightmaps;
  if (src->colour_replaces) gll_append((gll_t **) &dest->colour_replaces, (gll_t *) src->colour_replaces);
  if (src->force_colours) dest->force_colours = 1;
  if (src->entity_searches) gll_append((gll_t **) &dest->entity_searches, (gll_t *) src->entity_searches);
  if (src->entity_adds) gll_append((gll_t **) &dest->entity_adds, (gll_t *) src->entity_adds);
  if (src->entity_deletes) gll_append((gll_t **) &dest->entity_deletes, (gll_t *) src->entity_deletes);
  if (src->entity_edits) gll_append((gll_t **) &dest->entity_edits, (gll_t *) src->entity_edits);
  if (src->fd) dest->fd = src->fd;
  else if (src->outfile) dest->outfile = src->outfile;
  if (src->prefix) dest->prefix = src->prefix;
  if (src->suffix) dest->suffix = src->suffix;
  if (src->describe) dest->describe = src->describe;
  if (! dest->number) if (src->number) dest->number++;
  TRACE("finished\n");
}

void debug_options(option_t *option) {
  if (option->number) {
    DEBUG("option->number: %u\n", option->number);
  }
  if (option->strip_lightmaps) {
    DEBUG("option->strip_lightmaps: %u\n", option->strip_lightmaps);
  }
  if (option->force_colours) {
    DEBUG("option->force_colours: %s\n", option->force_colours);
  }
  if (option->colour_replaces) {
    DEBUG("option->colour_replaces: %p next: %p\n", option->colour_replaces, option->colour_replaces);
  }
  if (option->entity_searches) {
    DEBUG("option->entity_searches: %p next: %p\n", option->entity_searches, option->entity_searches);
  }
  if (option->entity_edits) {
    DEBUG("option->entity_edits: %p next: %p\n", option->entity_edits, option->entity_edits);
  }
  if (option->outfile) {
    DEBUG("option->outfile: %s\n", option->outfile);
  }
  if (option->prefix) {
    DEBUG("option->prefix: %s\n", option->prefix);
  }
  if (option->suffix) {
    DEBUG("option->suffix: %s\n", option->suffix);
  }
  if (option->describe) {
    DEBUG("option->describe: %u\n", option->describe);
  }
}
