#ifndef OPTION_H
#define OPTION_H
#include <stdio.h>
#include "data.h"

#define OPTION_NONE 0
#define OPTION_ARGUMENT 1
#define OPTION_FILENAME 2

#define EDIT_ADD 1
#define EDIT_EDIT 2
#define EDIT_DELETE 4

#define DESCRIBE_SUMMARY 1
#define DESCRIBE_STANDARD 2
#define DESCRIBE_ENTITIES 4
#define DESCRIBE_LIGHTS 8
#define DESCRIBE_CLASSES 16
#define DESCRIBE_ALL ~0

/*
  all the options that can be set

  if you do something like this:

    gsmhacking --make-swt flibble.gsm --classify wibble.gsm wobble.gsm

  you are making two option groups: { "make-swt" } and { "list"}
  each option group applies to all following gsm before the next option group
  so in effect the above is equivalent to:

    gsmhacking --make-swt flibble.gsm
    gsmhacking --classify wibble.gsm
    gsmhacking --classify wobble.gsm
*/

struct option {
  FILE *fd;
  char *outfile;
  char *prefix;
  char *suffix;
  unsigned int number;
  int scncount; /* how many SCNs use this option? */
  unsigned short describe;
  unsigned short strip_lightmaps;
  unsigned short force_colours;
  struct colour_replace_set *colour_replaces;
  struct entity_search_set *entity_searches;
  struct entity_edit_set *entity_edits;
  struct entity_add_set *entity_adds;
  struct entity_delete_set *entity_deletes;
};
typedef struct option option_t;

option_t **alloc_options(unsigned short);
option_t *alloc_option();
void cleanup_option(option_t *);
void initialise_option(option_t *);
void cleanup_options(unsigned short, option_t **);
struct scn **process_options(int *, char **, int *);
void copy_options(option_t *, option_t *);
int usage();
int help();
int parse_option(option_t *, int, char **, int);
int version();
void copy_options(option_t *, option_t *);
void debug_options(option_t *);

#endif
