#include "scnhacking.h"

pair_t **alloc_pairs(unsigned int n) {
  return (pair_t **) alloc_lots(n, "pair", sizeof(pair_t *), (void *(*)()) alloc_pair);
}

pair_t **alloc_more_pairs(unsigned int n, unsigned int num_pairs, pair_t **pairs) {
  return (pair_t **) alloc_more(n, "pair", sizeof(pair_t *), (void *(*)()) alloc_pair, num_pairs, (void **) pairs, sizeof(pair_t));
}

pair_t *alloc_pair() {
  return (pair_t *) alloc_one("pair", sizeof(pair_t));
}

void cleanup_pairs(unsigned int num_pairs, pair_t **pairs) {
  unsigned int i;

  if (! num_pairs) return;

  if (! pairs) {
    ERROR("no pairs!\n");
    return;
  }

  for (i = 0; i < num_pairs; i++) cleanup_pair(pairs[i]);
  TRACE2("freeing entity key/value pairs at %p\n", pairs);
  free(pairs);
}

void cleanup_pair(pair_t *pair) {
  if (! pair) {
    ERROR("no pair!\n");
    return;
  }
  if (pair->key_length) {
    if (pair->key) {
      TRACE2("freeing key at %p\n", pair->key);
      free(pair->key);
    }
    else {
      ERROR("no key in pair at %p (expected key of length %u)\n", pair, pair->key_length);
    }
  }
  if (pair->value_length) {
    if (pair->value) {
      TRACE2("freeing value at %p\n", pair->value);
      free(pair->value);
    }
    else {
      ERROR("no value in pair at %p (expected value of length %u)\n", pair, pair->value_length);
    }
  }
  TRACE2("freeing pair at %p\n", pair);
  free(pair);
}

/* -1 => not found, < -1 => error */
int get_value_for_pair(unsigned int num_pairs, pair_t **pairs, char *key) {
  unsigned int i;

  if (! pairs) {
    ERROR("no pairs\n");
    return -2;
  }

  for (i = 0; i < num_pairs; i++) {
    if (str_equiv(pairs[i]->key, key)) return (int) i;
  }

  TRACE("key %u not found in %u pairs at %p\n", key, num_pairs, pairs);
  return -1;
}

/* -1 => not found, < -1 => error */
int get_key_for_pair(unsigned int num_pairs, pair_t **pairs, char *value) {
  unsigned int i;

  if (! pairs) {
    ERROR("no pair\n");
    return -2;
  }

  for (i = 0; i < num_pairs; i++) {
    if (str_equiv(pairs[i]->value, value)) return (int) i;
  }

  TRACE("value %u not found in %u pairs at %p\n", value, num_pairs, pairs);
  return -1;
}
